#define		F_CPU		12000000UL
#define		SM		/* needed by avr/sleep.h */

#include <avr/cpufunc.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <inttypes.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "FreeRTOS.h"
#include "freertos_task.h"
#include "freertos_queue.h"

#include "avr_generic.h"
#include "avr_systick.h"

#define         SYSTICK0_BASE           0x08

/*****************************************************************************/


#define		IO_BASE_UART0		0x00
#define		IO_BASE_PORTOUT0	0x04
#define		IO_BASE_SYSTICK0	0x08

/* uart.h */

#define		UDR0		__IOR(IO_BASE_UART0+0x00)
#define		UCSRA0		__IOR(IO_BASE_UART0+0x01)
#define		UCSRB0		__IOR(IO_BASE_UART0+0x02)
#define		UBRR0		__IOR(IO_BASE_UART0+0x03)

/* UCSRA */
#define		RXB8		0
#define		PE		2
#define		DOR		3
#define		FE		4
#define		UDRE		5
#define		TXC		6
#define		RXC		7

/* UCSRB */
#define		TXB8		0
#define		UCSZ		1
#define		UPM0		2
#define		UPM1		3
#define		USBS		4
#define		UDRIE		5
#define		TXCIE		6
#define		RXCIE		7	

/* timer.h */

#define		TCNT0		__IOW(IO_BASE_TIMER0+0x00)
#define		TCR0		__IOR(IO_BASE_TIMER0+0x02)
#define		TSR0		__IOR(IO_BASE_TIMER0+0x03)

#define		TOF		7	/* timer overflow */
#define		TOFIE		7	/* timer overflow interrupt enable */
#define		TPRESC0		0	/* timer prescaler bit 0 */
#define		TPRESC1		1	/* timer prescaler bit 1 */

/* port.h */

#define		PORTOUT0	__IOR(IO_BASE_PORTOUT0+0x00)

/*****************************************************************************/

static inline void msleep_delay(uint16_t msec)
{
 while ( msec )
  {     _delay_loop_2((uint32_t)F_CPU/4000UL);
        msec--;
  }
}

void msleep(int t)
{
 vTaskDelay( t / portTICK_RATE_MS );
}

void main_task(void *arg)
{
 int	*iarg=arg;

 while ( 1 )
  {	PORTOUT0 ^= 1<<0;
	msleep(*iarg);
	
  }
}

void vApplicationIdleHook(void)
{
 sleep_cpu();
}

void vApplicationTickHook(void)
{
}

void vApplicationStackOverflowHook(void)
{
}

void vApplicationMallocFailedHook(void)
{
}

static QueueHandle_t queue_uart_tx;
static QueueHandle_t queue_uart_rx;

int uart_write_queue(void *cookie, const char *buff, size_t size)
{
 int	ssize;
 (void)cookie;

 ssize=0;
 while ( 0<size )
  {	xQueueSend(queue_uart_tx,buff,portMAX_DELAY); 
	UCSRB0 |= (1<<UDRIE);
	buff++;
	size--;
	ssize++;
  }
 return(ssize);
}

int uart_read_queue(void *cookie,char *buff,size_t size)
{
 int	ssize;
        
 ssize=0;
 while ( 0<size )
  {	BaseType_t	ret;
	ret=xQueueReceive(queue_uart_rx,buff,!ssize?portMAX_DELAY:0);
	buff++;
	size--;
	if ( ret )
		ssize++;
	else
		return(ssize);
  }
 return(ssize);
}

void uppercase_task(void * parameters)
{
 char   buff[8];
        
 while ( 1 )
  {     int     i,n;
        n=uart_read_queue(NULL,buff,8);
        for ( i=0; i<n; i++ )
         {      buff[i]=toupper(buff[i]);
         }
        uart_write_queue(NULL,buff,n);
  }
}

ISR(_VECTOR(3))
{

 BaseType_t xTaskWokenByReceive = pdFALSE;

 if ( UCSRA0 & (1<<UDRE) )
  {	BaseType_t	ret;
	char		c;

	ret=xQueueReceiveFromISR(queue_uart_tx,&c,&xTaskWokenByReceive);
	if ( ret )
		UDR0 = c;
	else
		UCSRB0 &= ~(1<<UDRIE);
  }

 if ( UCSRA0 & (1<<RXC) )
  {	BaseType_t	ret;
	char		c;

	c=(char)UDR0;
	ret=xQueueSendFromISR(queue_uart_rx,&c,&xTaskWokenByReceive);
	(void)ret;

	/*
	if ( ret == errQUEUE_FULL )
		data_lost_counter++;
	*/
  }

 if ( xTaskWokenByReceive != pdFALSE )
	taskYIELD();
}

int main(void)
{
 int	halfperiod;

 queue_uart_tx=xQueueCreate(8,1);
 queue_uart_rx=xQueueCreate(8,1);

 UBRR0 = 13-1;	 /* 115200 baud at 12000000MHz */
 UCSRB0 |= (1<<RXCIE);

 halfperiod=500;

 xTaskCreate(main_task,
        "main",
        configMINIMAL_STACK_SIZE,
        &halfperiod,
        1, 
        NULL );

 xTaskCreate(uppercase_task,
        "upper",
        configMINIMAL_STACK_SIZE,
        NULL,
        1, 
        NULL );

 vTaskStartScheduler();

 return(0);
}
